@@ -30,7 +30,6 @@ from . import testutils
 from .testutils import unittest, ConnectingTestCase, decorate_all_tests
 
 import psycopg2
-from psycopg2.extensions import b
 
 
 class TypesBasicTests(ConnectingTestCase):
@@ -69,13 +68,16 @@ class TypesBasicTests(ConnectingTestCase):
                         "wrong decimal quoting: " + str(s))
         s = self.execute("SELECT %s AS foo", (decimal.Decimal("NaN"),))
         self.assertTrue(str(s) == "NaN", "wrong decimal quoting: " + str(s))
-        self.assertTrue(type(s) == decimal.Decimal, "wrong decimal conversion: " + repr(s))
+        self.assertTrue(type(s) == decimal.Decimal,
+                        "wrong decimal conversion: " + repr(s))
         s = self.execute("SELECT %s AS foo", (decimal.Decimal("infinity"),))
         self.assertTrue(str(s) == "NaN", "wrong decimal quoting: " + str(s))
-        self.assertTrue(type(s) == decimal.Decimal, "wrong decimal conversion: " + repr(s))
+        self.assertTrue(type(s) == decimal.Decimal,
+                        "wrong decimal conversion: " + repr(s))
         s = self.execute("SELECT %s AS foo", (decimal.Decimal("-infinity"),))
         self.assertTrue(str(s) == "NaN", "wrong decimal quoting: " + str(s))
-        self.assertTrue(type(s) == decimal.Decimal, "wrong decimal conversion: " + repr(s))
+        self.assertTrue(type(s) == decimal.Decimal,
+                        "wrong decimal conversion: " + repr(s))
 
     def testFloatNan(self):
         try:
@@ -142,8 +144,8 @@ class TypesBasicTests(ConnectingTestCase):
             self.assertEqual(s, buf2.tobytes())
 
     def testArray(self):
-        s = self.execute("SELECT %s AS foo", ([[1,2],[3,4]],))
-        self.assertEqual(s, [[1,2],[3,4]])
+        s = self.execute("SELECT %s AS foo", ([[1, 2], [3, 4]],))
+        self.assertEqual(s, [[1, 2], [3, 4]])
         s = self.execute("SELECT %s AS foo", (['one', 'two', 'three'],))
         self.assertEqual(s, ['one', 'two', 'three'])
 
@@ -151,9 +153,12 @@ class TypesBasicTests(ConnectingTestCase):
         # ticket #42
         import datetime
         curs = self.conn.cursor()
-        curs.execute("create table array_test (id integer, col timestamp without time zone[])")
+        curs.execute(
+            "create table array_test "
+            "(id integer, col timestamp without time zone[])")
 
-        curs.execute("insert into array_test values (%s, %s)", (1, [datetime.date(2011,2,14)]))
+        curs.execute("insert into array_test values (%s, %s)",
+            (1, [datetime.date(2011, 2, 14)]))
         curs.execute("select col from array_test where id = 1")
         self.assertEqual(curs.fetchone()[0], [datetime.datetime(2011, 2, 14, 0, 0)])
 
@@ -161,12 +166,14 @@ class TypesBasicTests(ConnectingTestCase):
         curs.execute("select col from array_test where id = 2")
         self.assertEqual(curs.fetchone()[0], [])
 
-    def testEmptyArray(self):
+    def testEmptyArrayNoCast(self):
         s = self.execute("SELECT '{}' AS foo")
-        self.assertEqual(s, [])
-        s = self.execute("SELECT '{}'::text[] AS foo")
-        self.assertEqual(s, [])
+        self.assertEqual(s, '{}')
         s = self.execute("SELECT %s AS foo", ([],))
+        self.assertEqual(s, '{}')
+
+    def testEmptyArray(self):
+        s = self.execute("SELECT '{}'::text[] AS foo")
         self.assertEqual(s, [])
         s = self.execute("SELECT 1 != ALL(%s)", ([],))
         self.assertEqual(s, True)
@@ -190,7 +197,7 @@ class TypesBasicTests(ConnectingTestCase):
         ss = ['', '{', '{}}', '{' * 20 + '}' * 20]
         for s in ss:
             self.assertRaises(psycopg2.DataError,
-                psycopg2.extensions.STRINGARRAY, b(s), curs)
+                psycopg2.extensions.STRINGARRAY, s.encode('utf8'), curs)
 
     @testutils.skip_before_postgres(8, 2)
     def testArrayOfNulls(self):
@@ -209,23 +216,38 @@ class TypesBasicTests(ConnectingTestCase):
         curs.execute("insert into na (texta) values (%s)", ([None],))
         curs.execute("insert into na (texta) values (%s)", (['a', None],))
         curs.execute("insert into na (texta) values (%s)", ([None, None],))
-        curs.execute("insert into na (inta) values (%s)",  ([None],))
-        curs.execute("insert into na (inta) values (%s)",  ([42, None],))
-        curs.execute("insert into na (inta) values (%s)",  ([None, None],))
+        curs.execute("insert into na (inta) values (%s)", ([None],))
+        curs.execute("insert into na (inta) values (%s)", ([42, None],))
+        curs.execute("insert into na (inta) values (%s)", ([None, None],))
         curs.execute("insert into na (boola) values (%s)", ([None],))
         curs.execute("insert into na (boola) values (%s)", ([True, None],))
         curs.execute("insert into na (boola) values (%s)", ([None, None],))
 
-        # TODO: array of array of nulls are not supported yet
-        # curs.execute("insert into na (textaa) values (%s)", ([[None]],))
+        curs.execute("insert into na (textaa) values (%s)", ([[None]],))
         curs.execute("insert into na (textaa) values (%s)", ([['a', None]],))
-        # curs.execute("insert into na (textaa) values (%s)", ([[None, None]],))
-        # curs.execute("insert into na (intaa) values (%s)",  ([[None]],))
-        curs.execute("insert into na (intaa) values (%s)",  ([[42, None]],))
-        # curs.execute("insert into na (intaa) values (%s)",  ([[None, None]],))
-        # curs.execute("insert into na (boolaa) values (%s)", ([[None]],))
+        curs.execute("insert into na (textaa) values (%s)", ([[None, None]],))
+
+        curs.execute("insert into na (intaa) values (%s)",  ([[None]],))
+        curs.execute("insert into na (intaa) values (%s)", ([[42, None]],))
+        curs.execute("insert into na (intaa) values (%s)",  ([[None, None]],))
+
+        curs.execute("insert into na (boolaa) values (%s)", ([[None]],))
         curs.execute("insert into na (boolaa) values (%s)", ([[True, None]],))
-        # curs.execute("insert into na (boolaa) values (%s)", ([[None, None]],))
+        curs.execute("insert into na (boolaa) values (%s)", ([[None, None]],))
+
+    @testutils.skip_before_postgres(8, 2)
+    def testNestedArrays(self):
+        curs = self.conn.cursor()
+        for a in [
+            [[1]],
+            [[None]],
+            [[None, None, None]],
+            [[None, None], [1, None]],
+            [[None, None], [None, None]],
+            [[[None, None], [None, None]]],
+        ]:
+            curs.execute("select %s::int[]", (a,))
+            self.assertEqual(curs.fetchone()[0], a)
 
     @testutils.skip_from_python(3)
     def testTypeRoundtripBuffer(self):
@@ -324,30 +346,43 @@ class TypesBasicTests(ConnectingTestCase):
         self.assertEqual(1, l1)
 
     def testGenericArray(self):
-        a = self.execute("select '{1,2,3}'::int4[]")
-        self.assertEqual(a, [1,2,3])
-        a = self.execute("select array['a','b','''']::text[]")
-        self.assertEqual(a, ['a','b',"'"])
+        a = self.execute("select '{1, 2, 3}'::int4[]")
+        self.assertEqual(a, [1, 2, 3])
+        a = self.execute("select array['a', 'b', '''']::text[]")
+        self.assertEqual(a, ['a', 'b', "'"])
 
     @testutils.skip_before_postgres(8, 2)
     def testGenericArrayNull(self):
         def caster(s, cur):
-            if s is None: return "nada"
+            if s is None:
+                return "nada"
             return int(s) * 2
         base = psycopg2.extensions.new_type((23,), "INT4", caster)
         array = psycopg2.extensions.new_array_type((1007,), "INT4ARRAY", base)
 
         psycopg2.extensions.register_type(array, self.conn)
-        a = self.execute("select '{1,2,3}'::int4[]")
-        self.assertEqual(a, [2,4,6])
-        a = self.execute("select '{1,2,NULL}'::int4[]")
-        self.assertEqual(a, [2,4,'nada'])
+        a = self.execute("select '{1, 2, 3}'::int4[]")
+        self.assertEqual(a, [2, 4, 6])
+        a = self.execute("select '{1, 2, NULL}'::int4[]")
+        self.assertEqual(a, [2, 4, 'nada'])
+
+    @testutils.skip_before_postgres(8, 2)
+    def testNetworkArray(self):
+        # we don't know these types, but we know their arrays
+        a = self.execute("select '{192.168.0.1/24}'::inet[]")
+        self.assertEqual(a, ['192.168.0.1/24'])
+        a = self.execute("select '{192.168.0.0/24}'::cidr[]")
+        self.assertEqual(a, ['192.168.0.0/24'])
+        a = self.execute("select '{10:20:30:40:50:60}'::macaddr[]")
+        self.assertEqual(a, ['10:20:30:40:50:60'])
 
 
 class AdaptSubclassTest(unittest.TestCase):
     def test_adapt_subtype(self):
         from psycopg2.extensions import adapt
-        class Sub(str): pass
+
+        class Sub(str):
+            pass
         s1 = "hel'lo"
         s2 = Sub(s1)
         self.assertEqual(adapt(s1).getquoted(), adapt(s2).getquoted())
@@ -355,9 +390,14 @@ class AdaptSubclassTest(unittest.TestCase):
     def test_adapt_most_specific(self):
         from psycopg2.extensions import adapt, register_adapter, AsIs
 
-        class A(object): pass
-        class B(A): pass
-        class C(B): pass
+        class A(object):
+            pass
+
+        class B(A):
+            pass
+
+        class C(B):
+            pass
 
         register_adapter(A, lambda a: AsIs("a"))
         register_adapter(B, lambda b: AsIs("b"))
@@ -371,8 +411,11 @@ class AdaptSubclassTest(unittest.TestCase):
     def test_no_mro_no_joy(self):
         from psycopg2.extensions import adapt, register_adapter, AsIs
 
-        class A: pass
-        class B(A): pass
+        class A:
+            pass
+
+        class B(A):
+            pass
 
         register_adapter(A, lambda a: AsIs("a"))
         try:
@@ -384,8 +427,11 @@ class AdaptSubclassTest(unittest.TestCase):
     def test_adapt_subtype_3(self):
         from psycopg2.extensions import adapt, register_adapter, AsIs
 
-        class A: pass
-        class B(A): pass
+        class A:
+            pass
+
+        class B(A):
+            pass
 
         register_adapter(A, lambda a: AsIs("a"))
         try:
@@ -393,6 +439,19 @@ class AdaptSubclassTest(unittest.TestCase):
         finally:
             del psycopg2.extensions.adapters[A, psycopg2.extensions.ISQLQuote]
 
+    def test_conform_subclass_precedence(self):
+
+        import psycopg2.extensions as ext
+
+        class foo(tuple):
+            def __conform__(self, proto):
+                return self
+
+            def getquoted(self):
+                return 'bar'
+
+        self.assertEqual(ext.adapt(foo((1, 2, 3))).getquoted(), 'bar')
+
 
 class ByteaParserTest(unittest.TestCase):
     """Unit test for our bytea format parser."""
@@ -410,7 +469,7 @@ class ByteaParserTest(unittest.TestCase):
         expected.
         """
         import ctypes
-        lib = ctypes.cdll.LoadLibrary(psycopg2._psycopg.__file__)
+        lib = ctypes.pydll.LoadLibrary(psycopg2._psycopg.__file__)
         cast = lib.typecast_BINARY_cast
         cast.argtypes = [ctypes.c_char_p, ctypes.c_size_t, ctypes.py_object]
         cast.restype = ctypes.py_object
@@ -444,9 +503,10 @@ class ByteaParserTest(unittest.TestCase):
 
     def test_full_hex(self, upper=False):
         buf = ''.join(("%02x" % i) for i in range(256))
-        if upper: buf = buf.upper()
+        if upper:
+            buf = buf.upper()
         buf = '\\x' + buf
-        rv = self.cast(b(buf))
+        rv = self.cast(buf.encode('utf8'))
         if sys.version_info[0] < 3:
             self.assertEqual(rv, ''.join(map(chr, list(range(256)))))
         else:
@@ -457,7 +517,7 @@ class ByteaParserTest(unittest.TestCase):
 
     def test_full_escaped_octal(self):
         buf = ''.join(("\\%03o" % i) for i in range(256))
-        rv = self.cast(b(buf))
+        rv = self.cast(buf.encode('utf8'))
         if sys.version_info[0] < 3:
             self.assertEqual(rv, ''.join(map(chr, list(range(256)))))
         else:
@@ -469,7 +529,7 @@ class ByteaParserTest(unittest.TestCase):
         buf += string.ascii_letters
         buf += ''.join('\\' + c for c in string.ascii_letters)
         buf += '\\\\'
-        rv = self.cast(b(buf))
+        rv = self.cast(buf.encode('utf8'))
         if sys.version_info[0] < 3:
             tgt = ''.join(map(chr, list(range(32)))) \
                 + string.ascii_letters * 2 + '\\'
